import configparser
import datetime
import logging
import json
import io
import minio
import requests
import os
import pandas as pd
import sys
import urllib3


urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


# Configurables
WORKLOAD = 'workload'

NON_INFRA = 'NON_INFRA'
INFRA = 'INFRA'
VERSION = 'v1'
HEADERS = {
    'Content-Type': 'application/json',
    'X-ROLE-ENTERPRISE_ADMIN': '1',
    'x-nsx-roles': 'ZW50ZXJwcmlzZV9hZG1pbg=='
}
SSL_CERT_FILE = "SSL_CERT_FILE"
URL_ALL_FORMAT = 'https://%s:%s/api/v1/intelligence/workload/classifications'
IC_FEATURES_BUCKET = 'icfeatures'
PORT_FEATURES_LATEST = 'port_features_latest'
GRAPH_FEATURES_LATEST = 'graph_features_latest'
VM_CLASSIFICATIONS = ' \
{ \
    "results" : [ \
        { \
            "compute_id" : "zos-nonprod-hub-clone", \
            "classification" : { \
                "start_time" : 1001, \
                "last_modified_time" : 2001, \
                "type" : "NON_INFRA", \
                "confidence" : 0.999912440776825, \
                "version" : "v1", \
                "feedback" : "NON_INFRA", \
                "embeddings" : [ 1.0, 2.0, 3.0 ] \
            } \
        }, \
       { \
            "compute_id" : "some_random_compute_1234", \
            "classification" : { \
                "start_time" : 200000, \
                "last_modified_time" : 200001, \
                "type" : "NON_INFRA", \
                "confidence" : 50, \
                "version" : "v1", \
                "feedback" : "INFRA", \
                "embeddings" : [ 4.1, 5.2, 6.3 ] \
            } \
        } \
    ] \
}'


class FeedbackCollector(object):
    def __init__(self, workload_service, timedelta_hours=24):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._logger.setLevel(logging.INFO)
        self.all_workloads_url = URL_ALL_FORMAT % (workload_service, WORKLOAD_SERVICE_PORT)
        self.__feedback_filter = {
            "filters": [
                {
                    "key": "LAST_FEEDBACK_TIME",
                    "criteria": "BETWEEN",
                    "value": self.get_posix_timedelta_filter(timedelta_hours)
                }
            ]
        }
        self.timedelta_hours = timedelta_hours
        self._feedback_exists = None

    @property
    def feedback_exists(self, ):
        """
        Getter for a feedback_exists flag, which will be used before doing further processing
        :return: self._feedback_exists
        """
        return self._feedback_exists

    @staticmethod
    def get_posix_timedelta_filter(timedelta_hours=24):
        """
        Construct a posix time range  for `timedelta_hours`, e.g.  '1638400000-1638490000'

        :param int timedelta_hours: The numnber of hours to look into the past for new feedbacks
        :return str: A posix time range that we can pass as a filter to the workload api
        """
        now = datetime.datetime.now(datetime.timezone.utc)
        return "{}-{}".format(
            int((now-datetime.timedelta(hours=timedelta_hours)).timestamp()),
            int(now.timestamp())
        )

    def collect_feedback(self):
        """
        Query the workload pod for feedback on vm classifications over the `timedelta` hours, process it
        into a Dict where keys are the vm-identifier, the values are a Dict of original_classification
        and customer_feedback key-values, then process that into a pandas DF for further processing

        :return pandas.DataFrame | None: feedback results over the last `timedelta` period; if
            there is no feedback submitted by the customer within the `timedelta` period, returns None
        """
        feedback_response = self.request_feedback()
        if self._feedback_exists is True:
            res = self.feedback_vms_to_df(feedback_response)
        else:
            res = None
        return res

    def feedback_vms_to_df(self, feedback):
        """
        Process the Dict[Dict] payload of customer feedback into a Dataframe

        vm-id    original_classification   feedback_value
        vm1      INFRA                     NOT_INFRA
        ...

        :param feedback:
        :return: pandas.Dataframe object of vms to their feedback
        """
        # take the results from response, and parse into a pandas df with
        # columns of compute_id, feedback, feedback_last_update_timestamp
        # TODO: Make the dict keys constants so we only have to update in one place if things change in response
        row_dict = {}
        for classification in feedback["results"]:
            compute_id = classification["compute_id"]
            row_dict[compute_id] = self._build_row(classification)
        # index will be the compute_id col, which we don't need to spec in columns below
        df = pd.DataFrame.from_dict(
            row_dict,
            orient="index",
            columns=["feedback", "feedback_last_update_time"],
        )
        df.index.name = "compute_id"
        self._logger.info(
            "DF of feedback_vms in last 24 hours:\n %s", df.head()
        )
        self._logger.info("DF of feedback_vms has index: %s", df.index.name)
        return df

    def _build_row(self, classification_result):
        """
        Add the items from the classification of `classification_result` to the row for the
        respective compute/workload_id in the df that we are constructing

        :param Dict classification_result: a dict in the results from the workload api
        :return List[str] row: A List of feedback value, feedback timestamp
        """
        return [
            classification_result["classification"]["feedback"],
            classification_result["classification"]["feedback_last_update_time"]
        ]

    def request_feedback(self,):
        """
        Make the GET request to the workload pod for feedback in the last `timedelta` period

        :return: Deserialized payload of vm classifications with feedback in the past `timedelta` hours
        """
        # TODO: add some retry logic or switch to session + httpadapter
        self._logger.info('Requesting vms with feedback in last %s hours from %s',
                          self.timedelta_hours,
                          self.all_workloads_url)
        try:
            response = requests.post(
                self.all_workloads_url,
                data=json.dumps(self.__feedback_filter),
                headers=HEADERS,
                verify=os.environ.get(SSL_CERT_FILE),
                cert=("/cert/tls.crt", "/cert/tls.key"),
                timeout=5,
            )
            response.raise_for_status()
        except requests.exceptions.HTTPError as err:
            raise SystemExit(err)
        # deserialize response
        self._logger.info("Feedback workloads: %s", response.content)
        response = json.loads(response.text)
        if len(response['results']) > 0:
            self._logger.info(
                "Found %s workloads with feedback in the last %s hours: %s",
                len(response["results"]),
                self.timedelta_hours,
                response["results"]
            )
            self._feedback_exists = True
        else:
            self._logger.info(
                "Found %s workloads with feedback in the last %s hours. No further feedback"
                "processing to do until the customer provides feedback on classified workloads",
                len(response['results']),
                self.timedelta_hours,
            )
            self._feedback_exists = False
        return response


class FeedbackProcessor(object):
    def __init__(self, feedback_by_vm):
        """
        Take in a DF of vm's and their feedbacks, then query minio for the most recent orcs of
        Graph and port-based features. Uses the input DF as a mask, merges the Graph and Port
        features into a df of unfiltered features for the vms with feedback.

        :param Dataframe feedback_by_vm: DF of vms to their feedback values
        """
        self.feedback_by_vm = feedback_by_vm
        self._logger = logging.getLogger(self.__class__.__name__)
        self._logger.setLevel(logging.INFO)
        self._logger.info("Initializing minio client")

        self.minio_client = minio.Minio(
            "minio:9000",
            access_key=os.environ.get("AWS_ACCESS_KEY_ID"),
            secret_key=os.environ.get("AWS_SECRET_ACCESS_KEY"),
        )

    def check_icfeatures_bucket_exists(self):
        """
        Use the minio client to check if the IC_FEATURES_BUCKET exists. If not,
        exit with a message for the user to wait until IC has ran successfully.
        """
        # `icfeatures` bucket must exist before FeatureCreator runs
        if not self.minio_client.bucket_exists(IC_FEATURES_BUCKET):
            self._logger.info(
                "The bucket %s does not exist. This may mean that the feedback"
                " job is running before the bucket has been configured by the"
                " infraclassifier. Please wait for it to run before expecting"
                " feedback-reporting. The feedback job will now exit.",
                IC_FEATURES_BUCKET)
            sys.exit(0)
        else:
            self._logger.info(
                "Found bucket %s, querying for features of feedback vms",
                IC_FEATURES_BUCKET
            )

    def create_telemetry_payload(self):
        """
        Create a telemetry payload that is a dict of dicts, where the keys are vm identifiers,
        and the values are dicts containing feature names's as keys, and the values of those features
        as the values.

        :return: Dict[str, Dict[str,str]] dict of vm identifiers to dict of
            feature_name:feature_val pairs
        """
        feature_df = self.create_feedback_features_df()
        return self.feature_df_to_json(feature_df)

    def feature_df_to_json(self, feature_df):
        """
        Takes the df of combined Port+Graph features and writes it to a json payload
        :return str:
        """
        return json.dumps(feature_df.to_dict("index"))

    def create_feedback_features_df(self):
        """
        Creates combined graph and port features DF, which is filtered down to only
        the vms in `self.feedback_by_vms`

        :return: DF of features by vm
        """
        graph_feature_obj_dir = "{}.orc/".format(GRAPH_FEATURES_LATEST)
        port_features_obj_dir = "{}.orc/".format(PORT_FEATURES_LATEST)

        # the helper `merge_feature_df` merges `self._feedback_by_vm` df with the port/graph feats
        # each of the dfs below should be very wide, but short (assuming a small amount of
        # feedback in the last 24 hours)
        graph_features = self.merge_feature_df(self.get_features_df(graph_feature_obj_dir))
        port_features = self.merge_feature_df(self.get_features_df(port_features_obj_dir))

        # now take these two wide boys, and make 'em wider
        all_features = port_features.merge(
            graph_features, how="inner", left_index=True, right_index=True)
        return all_features

    def get_feature_keys(self, dir_object, bucket=IC_FEATURES_BUCKET):
        """
        The features get written to a directory object in minio, which contains many keys,
        where each key is a compressed partition of the orc. This method compiles all of the keys
        into a single list.

        :param str dir_object: The directory object in minio, e.g. `graph_features_latest.orc/`
        :param str bucket: the name of the bucket
        :return List[str]: the list of object keys
        """
        return [k.object_name for k in self.minio_client.list_objects(bucket, dir_object)]

    def get_features_df(self, feature_obj_name):
        """
        Get graph features from minio and convert to a Dataframe.

        :param str feature_obj_name: Either `graph_features_latest.orc/` or `port_features_latest.orc/`
        :return: DF of the features stored under the feature object dir specified by
            `feature_obj_name`
        """
        # check if it's there first
        self.check_icfeatures_bucket_exists()
        # TODO: convert feature_obj_name to an enum
        # we will build up df by concatenating temp_df's
        # use minio client to query bucket
        keys = self.get_feature_keys(feature_obj_name, bucket=IC_FEATURES_BUCKET)
        self._logger.debug("Found the following keys in the %s object: %s", feature_obj_name, keys)
        # Get features object data from minio
        df = pd.DataFrame()
        for key in keys:
            try:
                response = self.minio_client.get_object(IC_FEATURES_BUCKET, key)
                temp_df = pd.read_orc(io.BytesIO(response.data))
                df = pd.concat([df, temp_df])
            except Exception as e:
                self._logger.error('Encountered exception: %s', e)
            finally:
                response.close()
                response.release_conn()
        self._logger.debug(
            "DF constructed from %s in last 24 hours:\n %s \n with index name: %s",
            feature_obj_name, df.head(), df.index.name
        )
        self._logger.info("DF shape is now: %s", df.shape)
        return df

    def merge_feature_df(self, feature_df):
        """
        Merges `self.feedback_by_vm` and the df `feature_df`, on the inner join of the workload id

        :param pandas.Dataframe feature_df: dataframe of features
        :return pandas.DataFrame: `feature_df` but containing only the vms in `self.feedback_by_vm`
        """

        feature_df.rename(columns={"destination_vm": "compute_id"}, inplace=True)
        feature_df.set_index("compute_id", inplace=True)
        self._logger.info(
            "DF of feedback_vms in last 24 hours: %s", self.feedback_by_vm.head()
        )
        self._logger.info(
            "DF of features from the last 24 hours: %s", feature_df.head()
        )
        self._logger.info("VMS in index : \n %s", feature_df.index)
        feature_df = feature_df.merge(
            self.feedback_by_vm, how="inner", left_index=True, right_index=True
        )
        self._logger.info(
            "Features after inner join on feedback_vms in last 24 hours: %s", feature_df.head()
        )
        return feature_df


class TelemetryManager(object):
    def __init__(self, payload):
        """
        Manager class for posting the payload data to telemetry

        :param str payload: JSON formatted string of payload data
        """
        self._logger = logging.getLogger(self.__class__.__name__)
        self._logger.setLevel(logging.INFO)
        self.service_name = "telemetry"
        self.port = "443"
        self.base_url = "https://%s:%s/api/v1/platform/telemetry/"
        self.telemetry_url = self.base_url % (self.service_name, self.port)
        self.telemtry_ca_cert_path = os.environ.get("TELEMETRY_CA_CRT_PATH")
        self.client_cert = os.environ.get("SSL_CERT_FILE")
        self.keyfile = "/cert/tls.key"
        self._telemetry_definition = {"definitions": [
            {
                "aggregation_method": "SUM",
                "name": "infraclassifier_feedback",
                "value_type": "NUMBER"
            },
        ]}
        self.payload = {
            "telemetries": [
                {
                    "aggregation_method": "SUM",
                    "name": "infraclassifier_feedback",
                    "value": payload,
                    "value_type": "NUMBER"
                }
            ]
        }

    def define_feedback_telemetry(self):
        """
        Define a TelemetryDefinition that looks like `self._telemetry_definition

        :return bool: True if the telemetry is successfully defined, already defined, else False
        """
        definition_url = self.telemetry_url + "telemetry-definitions"
        try:
            response = requests.put(
                definition_url,
                json=self._telemetry_definition,
                verify=os.environ.get(SSL_CERT_FILE),
                cert=("/cert/tls.crt", self.keyfile),
                headers=HEADERS
            )
            response.raise_for_status()
        except requests.exceptions.HTTPError as err:
            raise SystemExit(err)
        self._logger.info(
            "Defined new telemetry for IC feedback with status: %s and content: %s",
            response.status_code,
            response.text
        )

    def update_feedback_telemetry(self):
        """
        Post feedback payload to telemetry via requests

        :param str data: The json encoded payload to send back to telemetry
        :return:
        """
        telemetries_url = self.telemetry_url + "telemetries"
        try:
            response = requests.put(
                telemetries_url,
                json=self.payload,
                verify=os.environ.get(SSL_CERT_FILE),
                cert=("/cert/tls.crt", self.keyfile),
                headers=HEADERS
            )
            response.raise_for_status()
        except requests.exceptions.HTTPError as err:
            raise SystemExit(err)
        self._logger.info(
            "Posted infraclassifier_feedback telemetry with status: %s, and content: %s",
            response.status_code,
            response.text
        )


# this function is duplicated for the time being and should be moved to a shared place
# this needs to be done to use the minio.Minio client
def set_up_certs(to_path="/opt/spark/work-dir/ca.crt", tls="/cert/tls.crt", ca="/cert/ca.crt"):
    """
    Appends the tls cert to the ca cert and copies it to a directory with write access, then exports the
    SSL_CERT_FILE env var needed by the python minio client

    :param to_path: the path to copy into
    :param tls: path to the tls.crt
    :param ca:  path to the ca.crt
    """
    # order matters here; write the ca, then the tls cert
    certs = [ca, tls]
    with open(to_path, "w") as outfile:
        for cert in certs:
            with open(cert, "r") as infile:
                outfile.write(infile.read())
    os.environ["SSL_CERT_FILE"] = to_path


def main():
    # set up logging
    logger = logging.getLogger("feedback_job")
    logger.setLevel(logging.INFO)
    set_up_certs()
    logger.info("Starting feedback job")
    # job_start_time = datetime.datetime.now(datetime.timezone.utc).timestamp()
    # job_start_time_pretty = datetime.datetime.now().strftime("%Y_%m_%d-%I:%M:%S_%p")
    # read config file for values like service_port, etc
    workload_service = WORKLOAD_SERVICE
    try:
        vms_with_feedback_pt24hr = FeedbackCollector(workload_service).collect_feedback()
        if vms_with_feedback_pt24hr is None:
            logger.info("There is no feedback present. No further work to do, job will now conclude.")
            sys.exit(0)
    except SystemExit as err:
        if err.code != 0:
            logger.error("Request to workload api failed with error: %s, the feedback job will now exit.", err)
            sys.exit(err.code)
        else:
            sys.exit(0)
    # TODO: Enable this once platform has fixed the telemetry service to handle features
    # telemetry_payload = FeedbackProcessor(vms_with_feedback_pt24hr).create_telemetry_payload()
    # instead, compute confusion metrics: FP, FN, TP, TN
    num_corrections = str(vms_with_feedback_pt24hr.count())
    try:
        telemetry_mgr = TelemetryManager(num_corrections)
        telemetry_mgr.define_feedback_telemetry()
        telemetry_mgr.update_feedback_telemetry()
    except SystemExit as err:
        if err.code != 0:
            logger.error(
                "Request to telemetry failed with error: %s, the feedback job will now exit.",
                err)
            sys.exit(err.code)
        else:
            sys.exit(0)

    logger.info("Feedback job will now conclude.")


if __name__ == "__main__":
    root_logger = logging.getLogger()
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    if root_logger.handlers:
        # or else, as I found out, we keep adding handlers and duplicate messages
        for handler in root_logger.handlers:
            if isinstance(handler, logging.StreamHandler):
                handler.setFormatter(
                    formatter
                )
                handler.setLevel(logging.INFO)
    else:
        ch = logging.StreamHandler(sys.stderr)
        ch.setLevel("INFO")
        ch.setFormatter(formatter)
        root_logger.addHandler(ch)
    config = configparser.ConfigParser()
    config.read("/opt/vmware/pace/config/infraclassifier.ini")
    WORKLOAD_SERVICE = config.get(WORKLOAD, "service_name")
    WORKLOAD_SERVICE_PORT = config.get(WORKLOAD, "service_port")
    main()
